import matplotlib as mpl
#mpl.use('MacOSX')
import matplotlib.pyplot as plot
import pylab
import networkx as netx
import serial
import time
from threading import Thread

#===============================================================================
# Constants
#===============================================================================
#-Discovery Information----------
NODE_TYPE = 0
DISTANCE_TRAVELED = 1

#-Response Types--------------
OK              = 'ok'
BRANCH_NODE     = 'bn'
LEFT_TURN_NODE  = 'ln'
RIGHT_TURN_NODE = 'rn'
END_NODE        = 'en'
STRAIGHT_NODE   = 'sn'
OBSTRUCTION     = 'ob'

#-Node Colors-----------------
STANDARD_NODE = '#99ccff'
ROBOT_NODE = '#FFFF66'
TARGET_NODE = '#FF5050'
UNDISCOVERED_NODE = '#325266'

#-Command types---------------
GO_STRAIGHT = 'go'
TURN_LEFT   = 'tl'
TURN_RIGHT  = 'tr'
TURN_AROUND = 'ta'

#-Compass Directions----------
NORTH = 0
SOUTH = 1
EAST  = 2
WEST  = 3

#===============================================================================
# Robot Data
#===============================================================================
robotDirection = NORTH
expectingDiscovery = False

#===============================================================================
# The Map
#===============================================================================
class Map(object):
    def __init__(self):
        self.newNodeID = 0
        self.robotNode = 0
        self.targetNode = 0
        self.nodeIDs = [0]
        self.nodePositions = [(0,0)]
        self.colors = [ROBOT_NODE]
        self.xPos = [0]
        self.yPos = [0]
        self.edges = []

        self.graph = netx.Graph()
        self.graph.add_node(self.newNodeID)

        mpl.rcParams['toolbar'] = 'None'

        fig = plot.figure()
        fig.canvas.set_window_title('Maple Console')
        plot.clf()

        netx.draw(self.graph, self.nodePositions, node_size = 500, node_color = self.colors)
        self.clickHandler = NodeClickFinder(self.xPos,self.yPos,self.nodeIDs,xtol=0.5,ytol=0.5)
        plot.connect('button_press_event', self.clickHandler)

        plot.show(block=False)

    def get_pos(self, nodeID):
        return (self.nodePositions[nodeID])

    def get_ID(self, x, y):
        for i in range(len(self.nodeIDs)):
            if (self.xPos[i] == x and self.yPos[i] == y):
                return self.nodeIDs[i]
        return -1

    def new_node(self, type):
        global robotDirection

        if (type == STRAIGHT_NODE):
            (x,y) = self.get_pos(self.robotNode)
            if (robotDirection == NORTH):
                y = y+1
            if (robotDirection == SOUTH):
                y = y-1
            if (robotDirection == EAST):
                x = x+1
            if (robotDirection == WEST):
                x = x-1
            # create new undiscovered node (straight)
            if self.get_ID(x,y) == -1:
                # node doesn't exist yet, make it undiscovered
                self.newNodeID += 1
                self.nodeIDs.append(self.newNodeID)
                self.nodePositions.append((x,y))
                self.xPos.append(x)
                self.yPos.append(y)
                self.colors.append(UNDISCOVERED_NODE)
                # add new information to the graph
                self.graph.add_node(self.newNodeID)
                self.edges.append((self.robotNode, self.newNodeID))
                self.graph.add_edge(self.newNodeID, self.robotNode, weight = 1)
            else:
                self.edges.append((self.robotNode, self.get_ID(x,y)))
                self.graph.add_edge(self.get_ID(x,y), self.robotNode, weight = 1)

        if (type == LEFT_TURN_NODE or type == BRANCH_NODE):
            (x,y) = self.get_pos(self.robotNode)
            if (turnLeft(robotDirection) == NORTH):
                y = y+1
            if (turnLeft(robotDirection) == SOUTH):
                y = y-1
            if (turnLeft(robotDirection) == EAST):
                x = x+1
            if (turnLeft(robotDirection) == WEST):
                x = x-1
            # create new undiscovered node (straight)
            if self.get_ID(x,y) == -1:
                # node doesn't exist yet, make it undiscovered
                self.newNodeID += 1
                self.nodeIDs.append(self.newNodeID)
                self.nodePositions.append((x,y))
                self.xPos.append(x)
                self.yPos.append(y)
                self.colors.append(UNDISCOVERED_NODE)
                # add new information to the graph
                self.graph.add_node(self.newNodeID)
                self.edges.append((self.robotNode, self.newNodeID))
                self.graph.add_edge(self.newNodeID, self.robotNode, weight = 1)
            else:
                self.edges.append((self.robotNode, self.get_ID(x,y)))
                self.graph.add_edge(self.get_ID(x,y), self.robotNode, weight = 1)

        if (type == RIGHT_TURN_NODE or type == BRANCH_NODE):
            (x,y) = self.get_pos(self.robotNode)
            if (turnRight(robotDirection) == NORTH):
                y = y+1
            if (turnRight(robotDirection) == SOUTH):
                y = y-1
            if (turnRight(robotDirection) == EAST):
                x = x+1
            if (turnRight(robotDirection) == WEST):
                x = x-1
            # create new undiscovered node (straight)
            if self.get_ID(x,y) == -1:
                # node doesn't exist yet, make it undiscovered
                self.newNodeID += 1
                self.nodeIDs.append(self.newNodeID)
                self.nodePositions.append((x,y))
                self.xPos.append(x)
                self.yPos.append(y)
                self.colors.append(UNDISCOVERED_NODE)
                # add new information to the graph
                self.graph.add_node(self.newNodeID)
                self.edges.append((self.robotNode, self.newNodeID))
                self.graph.add_edge(self.newNodeID, self.robotNode, weight = 1)
            else:
                self.edges.append((self.robotNode, self.get_ID(x,y)))
                self.graph.add_edge(self.get_ID(x,y), self.robotNode, weight = 1)

        self.update_graphics()

    def update_robot(self):
        global robotDirection
        (x,y) = self.get_pos(self.robotNode)
        if (robotDirection == NORTH):
            y += 1
        if (robotDirection == SOUTH):
            y -= 1
        if (robotDirection == EAST):
            x += 1
        if (robotDirection == WEST):
            x -= 1
        self.colors[self.robotNode] = STANDARD_NODE
        self.robotNode = self.get_ID(x,y)
        self.colors[self.robotNode] = ROBOT_NODE
        self.update_graphics()


    def get_path(self, x, y):
        global expectingDiscovery
        nodeID = self.get_ID(x,y)
        if (nodeID != -1):
            if self.colors[nodeID] == UNDISCOVERED_NODE:
                expectingDiscovery = True
            self.targetNode = nodeID
            self.colors[self.targetNode] = TARGET_NODE
            self.update_graphics()
            return netx.astar_path(self.graph, self.robotNode, self.targetNode)

    # Updates the display of the map
    def update_graphics(self):
        plot.clf()
        self.clickHandler.update(self.xPos, self.yPos, self.nodeIDs)
        netx.draw(self.graph, self.nodePositions, node_size = 500, node_color = self.colors)
        plot.draw() # bottleneck

#===============================================================================
# The click handler for the map
#===============================================================================
class NodeClickFinder:
    def __init__(self, xdata, ydata, IDs, xtol=None, ytol=None):
        self.data = zip(xdata, ydata, IDs)
        if xtol is None: xtol = ((max(xdata) - min(xdata))/float(len(xdata)))/2
        if ytol is None: ytol = ((max(ydata) - min(ydata))/float(len(ydata)))/2
        self.xtol = xtol
        self.ytol = ytol
        self.axis = pylab.gca()

    def update(self, xdata, ydata, IDs):
        self.data = zip(xdata, ydata, IDs)
        self.axis = pylab.gca()

    def __call__(self, event):
        global theMap
        if event.inaxes:
            clickX = event.xdata
            clickY = event.ydata
            if self.axis is None or self.axis == event.inaxes:
                IDs = []
                for x,y,a in self.data:
                    if (clickX - self.xtol < x < clickX + self.xtol and
                            clickY - self.ytol < y < clickY + self.ytol):
                        dx, dy = x-clickX, y-clickY
                        IDs += [(dx*dx+dy*dy, x, y, a)]
                if IDs:
                    IDs.sort() # to select the nearest node
                    distance, x, y, annote = IDs[0]
                    path = theMap.get_path(x, y)
                    command_robot(path)


#===============================================================================
# Direction Modification Functions
#===============================================================================
def directionToText(direction):
    if direction == NORTH:
        return 'North'
    if direction == SOUTH:
        return 'South'
    if direction == EAST:
        return 'East'
    if direction == WEST:
        return 'West'

def reverseDirection(direction):
    if direction == NORTH:
        return SOUTH
    elif direction == SOUTH:
        return NORTH
    elif direction == EAST:
        return WEST
    elif direction == WEST:
        return EAST

def turnLeft(direction):
    if direction == NORTH:
        return WEST
    elif direction == SOUTH:
        return EAST
    elif direction == EAST:
        return NORTH
    elif direction == WEST:
        return SOUTH

def turnRight(direction):
    if direction == NORTH:
        return EAST
    elif direction == SOUTH:
        return WEST
    elif direction == EAST:
        return SOUTH
    elif direction == WEST:
        return NORTH

#===============================================================================
# Simulate Robot travelling to destination node
#===============================================================================
def command_robot(path):
    global theMap
    global robotDirection
    global expectingDiscovery

    turns = turns_in_path(path)

    while (len(turns) > 0):
        instruction = turns.pop(0)
        response = send_serial(instruction)

        if (instruction == TURN_AROUND):
            if response == OK:
                robotDirection = reverseDirection(robotDirection)
            else:
                print 'failed acknowledgment'
                break
        elif (instruction == TURN_LEFT):
            if response == OK:
                robotDirection = turnLeft(robotDirection)
            else:
                print 'failed acknowledgment'
                break
        elif (instruction == TURN_RIGHT):
            if response == OK:
                robotDirection = turnRight(robotDirection)
            else:
                print 'failed acknowledgment'
                break
        elif (instruction == GO_STRAIGHT):
            #time.sleep(1)
            if response in [BRANCH_NODE, LEFT_TURN_NODE, RIGHT_TURN_NODE, END_NODE, STRAIGHT_NODE]:
                theMap.update_robot()
                #time.sleep(0.5)
                if (len(turns) == 0 and expectingDiscovery == True):
                    expectingDiscovery = False
                    theMap.new_node(response)
            else:
                print 'bad response'
                break

def send_serial(inst):
    ser = serial.Serial('/dev/tty.usbserial-A603UZ45', 9600)
    print 'Sending', inst
    inst = inst + '\n'
    cmd = inst.encode('utf-8')
    ser.write(cmd)
    print 'Waiting for response... '
    asdf = ser.readline()
    response = ser.readline().strip()
    print 'Response:', response
    return response
    # time.sleep(0.5)
    # if inst == TURN_AROUND or inst == TURN_LEFT or inst == TURN_RIGHT:
    #     return OK
    # if inst == GO_STRAIGHT:
    #     return BRANCH_NODE

#===============================================================================
# Calculate what turns are necessary to get to the target node
#===============================================================================
def turns_in_path(path):
    global robotDirection
    global theMap

    turns = []
    direction = robotDirection

    i = 0

    for nodeID in path[:-1]:
        destNode = path[i+1]
        x_dif = theMap.get_pos(destNode)[0] - theMap.get_pos(nodeID)[0]
        y_dif = theMap.get_pos(destNode)[1] - theMap.get_pos(nodeID)[1]

        if direction == NORTH:
            if x_dif > 0:
                turns += [TURN_RIGHT]
                direction = EAST
            elif x_dif < 0:
                turns += [TURN_LEFT]
                direction = WEST
            elif y_dif < 0:
                turns += [TURN_AROUND]
                direction = SOUTH
            turns += [GO_STRAIGHT]
        elif direction == SOUTH:
            if x_dif > 0:
                turns += [TURN_LEFT]
                direction = EAST
            elif x_dif < 0:
                turns += [TURN_RIGHT]
                direction = WEST
            elif y_dif > 0:
                turns += [TURN_AROUND]
                direction = NORTH
            turns += [GO_STRAIGHT]
        elif direction == EAST:
            if y_dif > 0:
                turns += [TURN_LEFT]
                direction = NORTH
            elif y_dif < 0:
                turns += [TURN_RIGHT]
                direction = SOUTH
            elif x_dif < 0:
                turns += [TURN_AROUND]
                direction = WEST
            turns += [GO_STRAIGHT]
        else:
            if y_dif > 0:
                turns += [TURN_RIGHT]
                direction = NORTH
            elif y_dif < 0:
                turns += [TURN_LEFT]
                direction = SOUTH
            elif x_dif > 0:
                turns += [TURN_AROUND]
                direction = EAST
            turns += [GO_STRAIGHT]

        i += 1

    return turns


#===============================================================================
# Main Execution Procedures
#===============================================================================
theMap = Map()
theMap.new_node(STRAIGHT_NODE)
theMap.update_robot()
theMap.new_node(BRANCH_NODE)
robotDirection = turnLeft(robotDirection)
theMap.update_robot()
theMap.new_node(RIGHT_TURN_NODE)
robotDirection = turnRight(robotDirection)
theMap.update_robot()
theMap.new_node(BRANCH_NODE)
robotDirection = reverseDirection(robotDirection)
theMap.update_robot()
robotDirection = turnLeft(robotDirection)
theMap.update_robot()
theMap.update_robot()
theMap.new_node(RIGHT_TURN_NODE)

# time.sleep(.2)
# theMap.update_robot()
# theMap.new_node(RIGHT_TURN_NODE)
# time.sleep(.2)
# robotDirection = EAST
# theMap.update_robot()
# theMap.new_node(BRANCH_NODE)
# time.sleep(.2)
# robotDirection = NORTH
# theMap.update_robot()
# theMap.new_node(RIGHT_TURN_NODE)
# time.sleep(.2)
plot.show()
