#define PIPE_ADDRESS_0 0xb4b5b600

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdio.h>
#include "uart.h"
#include "wireless.h"

#define F_CPU 16000000UL


//==============================================================================
// Macros
//==============================================================================

// Motor Pins
#define LEFT_MOTOR  4 // pin D6
#define RIGHT_MOTOR 5 // pin D7

// QTI Pins
#define QTI_LEFT       (PINC>>0 & 0b1) // pin C0
#define QTI_RIGHT      (PINC>>1 & 0b1) // pin C1
#define QTI_LEFT_WING  (PINC>>2 & 0b1) // pin C2
#define QTI_RIGHT_WING (PINC>>3 & 0b1) // pin C3

//LED pins
#define BLUE_LED  2
#define LEFT_LED  6
#define RIGHT_LED 7

// QTI Data
#define BLACK 0b1
#define WHITE 0b0

//Line Data
#define LEFT 0
#define RIGHT 1

// Node Types
#define BRANCH_NODE 0b00100000
#define LEFT_TURN   0b01000000
#define RIGHT_TURN  0b01100000

// Instructions
#define GO 0x01
#define TURN_LEFT   0x02
#define TURN_RIGHT  0x03
#define TURN_AROUND 0x04
#define STOP        0x05

// Pulse-Width for Wheel Spins (in 0.05 ms)
#define CLKWISE_FAST         62 //50
#define CLKWISE_SLOW         89 //71
#define NEUTRAL              94 //75
#define COUNTER_CLKWISE_SLOW 99 //79
#define COUNTER_CLKWISE_FAST 125//100

#define TURN_TIME 21000

#define left_direction OCR1BL
#define right_direction OCR1AL

//==============================================================================
// Global Variables
//==============================================================================

// UART file descriptor
// putchar and getchar are in uart.c
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

unsigned char command_number;
char last_response;

char onSide;

struct Packet {
  char data;
  char num;
};

char Ain;     // raw A to D number

//==============================================================================
// Turning Functions
//==============================================================================
void go_straight(void) {
  left_direction = COUNTER_CLKWISE_FAST;
  right_direction = CLKWISE_FAST;
  _delay_ms(1000);
  left_direction  = NEUTRAL;
  right_direction = NEUTRAL;
}

void turn_left(void) {
  left_direction  = CLKWISE_FAST;
  right_direction = CLKWISE_FAST;
  _delay_ms(500);
  left_direction  = NEUTRAL;
  right_direction = NEUTRAL;
  go_straight();
}

void turn_right(void) {
  left_direction  = COUNTER_CLKWISE_FAST;
  right_direction = COUNTER_CLKWISE_FAST;
  _delay_ms(500);
  left_direction  = NEUTRAL;
  right_direction = NEUTRAL;
  go_straight();
}

void turn_around(void) {
  left_direction  = CLKWISE_FAST;
  right_direction = CLKWISE_FAST;
  _delay_ms(1000);
   left_direction  = NEUTRAL;
  right_direction = NEUTRAL;
  go_straight();
}


//==============================================================================
// Line Following Function
//==============================================================================
/*
 * Follows the line. Stops and returns as soon as it
 * finds a node.
 */

void update_qtis(char* left, char* right){

  *left = ADCH;
  ADMUX |= (1<<MUX0);
  ADCSRA |= (1<<ADSC);
  printf("QTI left: %d, ",*left);

  *right = ADCH;
  ADMUX &= ~(1<<MUX0);
  ADCSRA |= (1<<ADSC);
  printf("QTI right: %d\n\r",*right);
}

char follow_line(void) {
  char node_found = 0;
  char qti_left;
  char qti_right;
  float error;
  int difference;

  float err_sum = 0;
  float err_diff = 0;
  float prev_err = 0;

  //float P = 0.16;
  float m = 0.0308;
  float b = -3.31;

  while (1) {

    update_qtis(&qti_left, &qti_right);

  	if (qti_left < qti_right && qti_right > 215) onSide = LEFT;
  	if (qti_left > qti_right && qti_left > 215) onSide = RIGHT;

  	// Going off left side
  	if (qti_left < 181 && qti_right < 233 && onSide == LEFT) {
  		error = -4;
  	}
  	// Going off right side
  	else if (qti_left < 236 && qti_right < 140 && onSide == RIGHT) {
  		error = 4;
  	}
  	// Cruising down the middle
  	else {
  		difference = qti_left-qti_right+100;
  		error = (m*((float)difference) + b);
  	}
  	err_sum += error;
  	err_diff = error - prev_err;
  	prev_err = error;

  	right_direction = (char)(87 - 0.08*error - 2.5*err_diff);
  	left_direction = (char)(101 - 0.08*error - 2.5*err_diff);

    if (QTI_LEFT_WING == BLACK) {
      left_direction = COUNTER_CLKWISE_SLOW;
      right_direction = CLKWISE_SLOW;
      _delay_ms(200);
      left_direction = NEUTRAL;
      right_direction = NEUTRAL;
      PORTD |= (1 << LEFT_LED) | (1 << RIGHT_LED);
      if (QTI_RIGHT_WING == BLACK)
        return BRANCH_NODE;
      else
      	return LEFT_TURN;
    }
	  else if (QTI_RIGHT_WING == BLACK) {
      left_direction = COUNTER_CLKWISE_SLOW;
      right_direction = CLKWISE_SLOW;
	    _delay_ms(200);
	    left_direction = NEUTRAL;
	    right_direction = NEUTRAL;
      PORTD |= (1 << LEFT_LED) | (1 << RIGHT_LED);
	    if (QTI_LEFT_WING == BLACK)
	  	  return BRANCH_NODE;
	    else
      	return RIGHT_TURN;
    }

    if (qti_left < 200) {
      PORTD |= (1<<LEFT_LED);
    }
	  else {
	    PORTD &= ~(1<<LEFT_LED);
	  }
    if (qti_right < 200) {
      PORTD |= (1<<RIGHT_LED);
    }
	  else {
      PORTD &= ~(1<<RIGHT_LED);
    }

  }
}



//==============================================================================
// Communication Functions
//==============================================================================

char receive_inst_from_base(void) {
  unsigned char pipe_received;
  char transmission_result;
  struct Packet packet;

  do{
    set_mode(RX_MODE);
    fprintf(stdout, "Switched to RX\n");
    while(RX_FIFO_status() != 1);
    pipe_received = read_RX_payload((char*) &packet);
    fprintf(stdout,"Expecting command number: %d, Received: %d\n", command_number, packet.num);
    if (packet.num != command_number) {
      set_mode(TX_MODE);
      fprintf(stdout, "Responding to duplicate\n");
      packet.data = last_response;
      transmission_result = transmit_packet((char*) &packet);
      if (transmission_result == 2) fprintf(stdout,"transmission failed\n");
      else fprintf(stdout, "Resent response\n");
    }
  } while(packet.num != command_number);

  fprintf(stdout, "New command received: %d\n", packet.data);
  return packet.data;

}

void report_to_base(char response) {
  unsigned char pipe_received;
  char transmission_result;
  struct Packet packet;

  packet.num = command_number;
  packet.data = response;
  set_mode(TX_MODE);

  fprintf(stdout, "Switched to TX\n");
  transmission_result = transmit_packet((char*) &packet);
  if (transmission_result == 2) fprintf(stdout,"transmission failed\n");
  else fprintf(stdout, "Sent response\n");

  last_response = packet.data;
  command_number++;

}

void connect_to_base(void) {
  char connectionResult;

  while(1){
    connectionResult = receive_inst_from_base();
    if (connectionResult == 0x42) {
      PORTD &= ~(1<<2);
      report_to_base(connectionResult);
      return;
    }
  }
}


//==============================================================================
// Initialization
//==============================================================================
void initialize(void) {
  // Initialize the ADC and start a conversion
  ADMUX = (1<<ADLAR) | (1<<REFS0);
  ADCSRA = (1<<ADEN) | (1<<ADSC) + 7;

  // Initialize the UART
  uart_init();
  stdout = stdin = stderr = &uart_str;
  fprintf(stdout,"\n\n\n\n\n\n\n\n\n\n_____Welcome to Project Maple_____\n\r");

  // PORT D is an input except for UART on D.1 and the motors
  DDRD = (1<<1) | (1<<LEFT_MOTOR) | (1<<RIGHT_MOTOR);

  // Set up LED on D.2, QTI LEDs on D.4,D.5
  DDRD |= (1<<BLUE_LED)|(1<<LEFT_LED)|(1<<RIGHT_LED);

  // Pullups
  PORTD |= (1<<BLUE_LED)|(1<<LEFT_LED)|(1<<RIGHT_LED);

  PORTA |= (1<<0) | (1<<1);

  OCR1AH = 0;
  OCR1AL = NEUTRAL;
  OCR1BH = 0;
  OCR1BL = NEUTRAL;

  TCCR1B = (1<<CS12) | (1<<WGM12);
  TCCR1A |= (1<<WGM10) | (1<<WGM11);
  TCCR1A |= (1<<COM1A1) | (1<<COM1B1);

  // pull up resistors on the QTI sensors (pins C0,C1,C2)
  PORTC |= (0b1 << 3) | (0b1 << 2) | (0b1 << 1) | (0b1 << 0);

  left_direction = NEUTRAL;
  right_direction = NEUTRAL;

  //radio init
  PAYLOAD_LENGTH = 2; //set the payload length of each packet
  init_RF(0); //disable enhanced shockburst
  set_RX_addr(PIPE_ADDRESS_0, 0);
  set_TX_addr(PIPE_ADDRESS_0);

  //base station comm
  command_number = 0;
  last_response = STOP;
}

void blink(char led){
  PORTD &= ~(1 << led);
  _delay_ms(500);
  PORTD |= (1 << led);
  _delay_ms(500);
  PORTD &= ~(1 << led);
  _delay_ms(500);
  PORTD |= (1 << led);
  _delay_ms(1000);
}

void blink_two(char led1, char led2){
  PORTD &= ~(1 << led1);
  PORTD &= ~(1 << led2);
  _delay_ms(500);
  PORTD |= (1 << led1);
  PORTD |= (1 << led2);
  _delay_ms(500);
  PORTD &= ~(1 << led1);
  PORTD &= ~(1 << led2);
  _delay_ms(500);
  PORTD |= (1 << led1);
  PORTD |= (1 << led2);
  _delay_ms(1000);
}

//==============================================================================
// Main Execution Function
//==============================================================================
int main(void) {
  char newInst;
  char newNode;

  initialize();

  connect_to_base();

  while(1) {
    newInst = receive_inst_from_base();
    if (newInst == GO){
      newNode = follow_line();
      report_to_base(newNode);
    }
    else if (newInst == TURN_LEFT){
	  turn_left();
      blink(LEFT_LED);
      report_to_base(TURN_LEFT);
    }
    else if (newInst == TURN_RIGHT){
	  turn_right();
      blink(RIGHT_LED);
      report_to_base(TURN_RIGHT);
    }
	  else if (newInst == TURN_AROUND){
	    turn_around();
	    blink_two(LEFT_LED, RIGHT_LED);
	    report_to_base(TURN_AROUND);
	  }
    else {
	    PORTD &= ~(1 << LEFT_LED);
      PORTD &= ~(1 << RIGHT_LED);
	    fprintf(stdout, "Invalid Command Received\n");
    }
  }

}

